package main

import (
	"context"
	"encoding/json"
	"fmt"
	"net"
	"net/http"
	"strings"
	"time"

	"github.com/camunda-cloud/zeebe/clients/go/pkg/zbc"
	"github.com/google/uuid"
	"github.com/stripe/stripe-go/v72/webhookendpoint"
	cardpayment "gitlab.mashplace.com/mashplace/backend/cardpayment/api"
	"gitlab.mashplace.com/mashplace/backend/cardpayment/internal/model/mongo"
	"gitlab.mashplace.com/mashplace/backend/cardpayment/internal/service"
	"gitlab.mashplace.com/mashplace/backend/cardpayment/version"
	"gitlab.mashplace.com/mashplace/backend/common/api/health"
	"gitlab.mashplace.com/mashplace/backend/common/generic/addr"
	"gitlab.mashplace.com/mashplace/backend/common/generic/bpms"
	"gitlab.mashplace.com/mashplace/backend/common/generic/env"
	"gitlab.mashplace.com/mashplace/backend/common/generic/grpc"
	"gitlab.mashplace.com/mashplace/backend/common/generic/httpwrapper"
	logger "gitlab.mashplace.com/mashplace/backend/common/generic/log"
	"gitlab.mashplace.com/mashplace/backend/common/generic/metrics"
	"gitlab.mashplace.com/mashplace/backend/common/generic/tracer"
)

var (
	envMongoURL      = env.ResolveString("MONGO_URL", "mongodb:27017")
	envMongoUsername = env.ResolveString("MONGO_USERNAME", "cardpayment-data")
	envMongoPassword = env.ResolveString("MONGO_PASSWORD", "cardpayment-data")
	envMongoAuthDB   = env.ResolveString("MONGO_AUTH", "admin")
	envMongoDB       = env.ResolveString("MONGO_DATABASE", "cardpayment-data")
	envMongoSSLMode  = env.ResolveBool("MONGO_SSL_MODE", false)
)

var (
	envStripeSecretKey = env.ResolveString("STRIPE_SECRET_KEY", "sk_test_fwK7y4sYjzWGwuzNEg3aDv2y")

	envSiteURL          = env.ResolveString("SITE_URL", "https://dev.env.mashroom.com")
	envStripeSuccessURL = env.ResolveString(
		"STRIPE_CHECKOUT_SUCCESS_URL",
		envSiteURL+"/callback/stripe/success",
	)
	envStripeCancelURL = env.ResolveString(
		"STRIPE_CHECKOUT_CANCEL_URL",
		envSiteURL+"/callback/stripe/cancel",
	)
	envStripeWebhookURL = env.ResolveString("STRIPE_WEBHOOK_URL", envSiteURL+"/callback/stripe/webhooks")
	envInsuranceAccount = env.ResolveString("STRIPE_INSURANCE_ACCOUNT", "acct_1GVFKbBwNAF9fwRb")

	// WebhookSecret - you can find your webhook endpoint's secret in your webhook settings
	// in the Developer Dashboard
	envWebhookSecret = env.ResolveString("STRIPE_WEBHOOK_SECRET", "")
)

func createWebhookEndpoint(webhookURL string) (secret string, err error) {
	if webhookURL == "" {
		return
	}

	whList := webhookendpoint.List(&stripe.WebhookEndpointListParams{})

	if whList != nil {
		for _, wh := range whList.WebhookEndpointList().Data {
			if strings.Contains(wh.URL, webhookURL) {
				return wh.Secret, nil
			}
		}
	}

	wh, err := webhookendpoint.New(&stripe.WebhookEndpointParams{
		URL: stripe.String(webhookURL),
		EnabledEvents: []*string{
			stripe.String("checkout.session.completed"),
			stripe.String("checkout.session.async_payment_succeeded"),
			stripe.String("checkout.session.async_payment_failed"),
			stripe.String("mandate.updated"),
			stripe.String("payment_method.automatically_updated"),
			stripe.String("charge.succeeded"),
			stripe.String("charge.failed"),
			stripe.String("payment_intent.succeeded"),
			stripe.String("payment_intent.created"),
			stripe.String("payment_intent.processing"),
			stripe.String("payment_intent.payment_failed"),
		},
	})
	if err != nil {
		return
	}

	return wh.Secret, nil
}

func serverRequestCallback(
	srv service.Service,
) (grpc.ServerCallbackRequestFunc, grpc.ServerCallbackResponseFunc) {
	return func(ctx context.Context, req interface{}) {
			if !grpc.MethodIsMutation(ctx) {
				return
			}

			mdata := grpc.MethodMetadata(ctx)

			reqid := grpc.MethodRequestId(ctx)

			reqbs, _ := json.Marshal(req)

			var reqm map[string]interface{}

			_ = json.Unmarshal(reqbs, &reqm)

			fname := string(mdata.FullName())

			srv.RPCCallback(
				ctx,
				"grpc",
				"call",
				fname,
				reqid,
				reqm,
				nil,
			)
		}, func(ctx context.Context, resp interface{}, err error) {
			if err == nil {
				return
			}

			if !grpc.MethodIsMutation(ctx) {
				return
			}

			mdata := grpc.MethodMetadata(ctx)

			reqid := grpc.MethodRequestId(ctx)

			fname := string(mdata.FullName())

			srv.RPCCallback(
				ctx,
				"grpc",
				"error",
				fname,
				reqid,
				map[string]interface{}{"error": err.Error()},
				nil,
			)
		}
}

func main() {
	ctx := context.Background()

	defer tracer.Close()

	repository, err := mongo.NewRepository(ctx, &mongo.Config{
		URL:          envMongoURL,
		Username:     envMongoUsername,
		Password:     envMongoPassword,
		Database:     envMongoDB,
		AuthSource:   envMongoAuthDB,
		MongoSSLMode: envMongoSSLMode,
	})
	if err != nil {
		panic(err)
	}

	err = repository.Migrate(ctx)
	if err != nil {
		panic(err)
	}

	var srv service.Service

	tw := &httpwrapper.CallbackTransportWrapper{
		Transport: http.DefaultTransport,
		Callback: func(
			ctx context.Context,
			req *http.Request,
			resp *http.Response,
			reqdata, respdata map[string]interface{},
		) {
			if srv == nil {
				return
			}

			reqid := uuid.New().String()

			srv.RPCCallback(
				ctx,
				"integration",
				"request",
				req.URL.String(),
				reqid,
				reqdata,
				map[string]interface{}{
					"method": req.Method,
				},
			)

			srv.RPCCallback(
				ctx,
				"integration",
				"response",
				req.URL.String(),
				reqid,
				respdata,
				map[string]interface{}{
					"code": resp.StatusCode,
				},
			)
		},
	}

	httpClient := &http.Client{
		Timeout:   80 * time.Second,
		Transport: tw,
	}

	stripe.Key = envStripeSecretKey
	stripe.NewIdempotencyKey()

	stripe.SetHTTPClient(httpClient)
	stripe.SetBackend(stripe.APIBackend, stripe.GetBackendWithConfig(
		stripe.APIBackend,
		&stripe.BackendConfig{
			HTTPClient: httpClient,
		},
	))

	secret, err := createWebhookEndpoint(envStripeWebhookURL)
	if err != nil {
		panic(err)
	}

	if len(secret) > 0 {
		envWebhookSecret = secret
	}

	var zeebeClient zbc.Client

	if bpms.URL != "" {
		zeebeClient, err = zbc.NewClient(&zbc.ClientConfig{
			UsePlaintextConnection: true,
			GatewayAddress:         bpms.URL,
		})
		if err != nil {
			panic(err)
		}
	}

	srv, err = service.NewService(ctx, &service.Config{
		Repository:  repository,
		ZeebeClient: zeebeClient,
		Stripe: service.Stripe{
			Currency:           stripe.CurrencyGBP,
			InsuranceAccount:   envInsuranceAccount,
			CallbackSuccessURL: fmt.Sprintf("%s?session_id={CHECKOUT_SESSION_ID}", envStripeSuccessURL),
			CallbackCancelURL:  fmt.Sprintf("%s?session_id={CHECKOUT_SESSION_ID}", envStripeCancelURL),
			WebhookSecret:      envWebhookSecret,
		},
	})
	if err != nil {
		panic(err)
	}

	listener, err := net.Listen("tcp", addr.Grpc)
	if err != nil {
		panic(err)
	}

	grpcServer, err := grpc.NewServer("cardpayment", grpc.ServerCallback(
		serverRequestCallback(srv),
	))
	if err != nil {
		panic(err)
	}

	cardpayment.RegisterCardpaymentServiceServer(grpcServer, srv)

	logger.Infoln("cardpayment started")

	healthServer := version.NewHealthServer()
	health.RegisterHealthServiceServer(grpcServer, healthServer)

	go func() {
		_ = metrics.ConfigureListenAndServe(grpcServer)
	}()

	err = grpcServer.Serve(listener)
	if err != nil {
		logger.Errorln("gRPC error", err)
	}
}
