package model

import (
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/generic/util"

	"go.mongodb.org/mongo-driver/bson/primitive"

	cardpayment "gitlab.mashplace.com/mashplace/backend/cardpayment/api"
	extras "gitlab.mashplace.com/mashplace/backend/extras/api"
)

// Confirm represents admin confirmation
type Confirm struct {
	Admin       string    `bson:"admin"`
	Confirmed   bool      `bson:"confirmed"`
	ConfirmedAt time.Time `bson:"confirmed_at"`
}

// Service represents extras service
type Service struct {
	Created       time.Time          `bson:"created"`
	Updated       time.Time          `bson:"updated"`
	ID            primitive.ObjectID `bson:"_id"`
	ItemID        string             `bson:"item_id"`
	ServiceID     string             `bson:"service_id"`
	TransferID    string             `bson:"transfer_id"`
	Code          string             `bson:"code"`
	Title         string             `bson:"title"`
	Description   string             `bson:"description"`
	Price         float64            `bson:"price"`
	TaxAmount     float64            `bson:"tax_amount"`
	TaxPercentage float64            `bson:"tax_percentage"`
	Confirm       Confirm            `bson:"confirm"`
	Type          extras.ServiceType `bson:"type"`
	TaxType       extras.TaxType     `bson:"tax_type"`
	Role          extras.ServiceRole `bson:"role"`
	Primary       bool               `bson:"primary"`
}

// Charge represents stripe-provided charge
type Charge struct {
	Items          []*Service                `bson:"items,omitempty"`
	ID             string                    `bson:"_id,omitempty"`
	ClientSecret   string                    `bson:"client_secret,omitempty"`
	UserID         string                    `bson:"user_id,omitempty"`
	UserEmail      string                    `bson:"user_email,omitempty"`
	Property       string                    `bson:"property,omitempty"`
	NotifyID       string                    `bson:"notify_id"`
	NotifyAddress  string                    `bson:"notify_address"`
	InvoiceNumber  string                    `bson:"invoice_number"`
	InvoiceID      string                    `bson:"invoice_id"`
	SessionID      string                    `bson:"session_id,omitempty"`
	SubscriptionID string                    `bson:"subscription_id,omitempty"`
	Total          int64                     `bson:"total,omitempty"`
	Type           cardpayment.PaymentType   `bson:"type"`
	Status         cardpayment.PaymentStatus `bson:"status,omitempty"`
	CreatedAt      time.Time                 `bson:"created_at,omitempty"`
	UpdatedAt      time.Time                 `bson:"updated_at,omitempty"`
	Transferred    bool                      `bson:"transferred"`
}

// RenderCharge converts model.Charge into gRPC API data.
func (m *Charge) RenderCharge() *cardpayment.Charge {
	if m == nil {
		return &cardpayment.Charge{}
	}

	c := &cardpayment.Charge{
		Id:            m.ID,
		ClientSecret:  m.ClientSecret,
		UserId:        m.UserID,
		UserEmail:     m.UserEmail,
		Property:      m.Property,
		NotifyId:      m.NotifyID,
		NotifyAddress: m.NotifyAddress,
		InvoiceNumber: m.InvoiceNumber,
		InvoiceId:     m.InvoiceID,
		SessionId:     m.SessionID,
		Total:         m.Total,
		Type:          m.Type,
		Status:        m.Status,
		Transferred:   m.Transferred,
		CreatedAt:     util.ToUnixTime(m.CreatedAt),
		UpdatedAt:     util.ToUnixTime(m.UpdatedAt),
	}

	var services []*cardpayment.Service
	for _, s := range m.Items {
		services = append(services, &cardpayment.Service{
			Id:            s.ID.Hex(),
			Primary:       s.Primary,
			ItemId:        s.ItemID,
			ServiceId:     s.ServiceID,
			TransferId:    s.TransferID,
			Code:          s.Code,
			Title:         s.Title,
			Description:   s.Description,
			Price:         s.Price,
			TaxAmount:     s.TaxAmount,
			TaxPercentage: s.TaxPercentage,
			Type:          s.Type,
			TaxType:       s.TaxType,
			Role:          s.Role,
			Created:       util.ToUnixTime(s.Created),
			Updated:       util.ToUnixTime(s.Updated),
			Confirm: &cardpayment.Confirm{
				Admin:       s.Confirm.Admin,
				Confirmed:   s.Confirm.Confirmed,
				ConfirmedAt: util.ToUnixTime(s.Confirm.ConfirmedAt),
			},
		})
	}

	c.Items = services

	return c
}
