// Package mongo provides mongo repository implementation for cardpayment service
package mongo

import (
	"context"
	"crypto/tls"
	"fmt"
	"time"

	"go.mongodb.org/mongo-driver/bson"
	"go.mongodb.org/mongo-driver/bson/primitive"
	"go.mongodb.org/mongo-driver/mongo"
	"go.mongodb.org/mongo-driver/mongo/options"

	"gitlab.mashplace.com/mashplace/backend/cardpayment/internal/model"
	v1 "gitlab.mashplace.com/mashplace/backend/cardpayment/internal/model/mongo/migrations/v1"
	"gitlab.mashplace.com/mashplace/backend/common/api/generic/query"
	"gitlab.mashplace.com/mashplace/backend/common/generic/migrations"
	mongoMigrations "gitlab.mashplace.com/mashplace/backend/common/generic/migrations/mongo"
	"gitlab.mashplace.com/mashplace/backend/common/generic/mongoquery"
)

const (
	chargesCollection    = "charges"
	mandatesCollection   = "mandates"
	userOffersCollection = "user_offers"
	changelogCollection  = "changelog"
)

// Config mongo repository config
type Config struct {
	URL          string
	Username     string
	Password     string
	Database     string
	AuthSource   string
	MongoSSLMode bool
}

type repositoryImpl struct {
	client               *mongo.Client
	database             *mongo.Database
	collectionCharges    *mongo.Collection
	collectionMandates   *mongo.Collection
	collectionUserOffers *mongo.Collection
	collectionChangelog  *mongo.Collection
}

func validateConfig(config *Config) error {
	if config == nil {
		return fmt.Errorf("invalid repository config")
	}

	if config.URL == "" {
		return fmt.Errorf("empty url in repository config")
	}

	if config.Username == "" {
		return fmt.Errorf("empty username in repository config")
	}

	if config.Password == "" {
		return fmt.Errorf("empty password in repository config")
	}

	if config.Database == "" {
		return fmt.Errorf("empty database in repository config")
	}

	if config.AuthSource == "" {
		return fmt.Errorf("empty database in repository config")
	}

	return nil
}

// NewRepository returns new repository instance
func NewRepository(ctx context.Context, config *Config) (model.Repository, error) {
	err := validateConfig(config)
	if err != nil {
		return nil, err
	}

	clientOptions := options.Client().ApplyURI(fmt.Sprintf("mongodb://%s", config.URL)).
		SetAuth(options.Credential{
			AuthSource:  config.AuthSource,
			Username:    config.Username,
			Password:    config.Password,
			PasswordSet: true,
		})

	if config.MongoSSLMode {
		clientOptions.SetTLSConfig(&tls.Config{})
	}

	client, err := mongo.NewClient(clientOptions)
	if err != nil {
		return nil, err
	}

	err = client.Connect(ctx)
	if err != nil {
		return nil, err
	}

	db := client.Database(config.Database)

	return &repositoryImpl{
		client:               client,
		database:             db,
		collectionCharges:    db.Collection(chargesCollection),
		collectionMandates:   db.Collection(mandatesCollection),
		collectionUserOffers: db.Collection(userOffersCollection),
		collectionChangelog:  db.Collection(changelogCollection),
	}, nil
}

func (repo *repositoryImpl) Migrate(ctx context.Context) error {
	return migrations.MigrateSchema(
		mongoMigrations.WithInstance(ctx, repo.database),
		&mongoMigrations.Migration{
			Version: "0",
			Up: func(ctx context.Context, db *mongo.Database, migrationID string) (int, error) {
				return -1, nil
			},
		},
		&mongoMigrations.Migration{
			Version: "1",
			Up:      v1.Up,
		},
	)
}

func (repo *repositoryImpl) ChargeSave(ctx context.Context, charge *model.Charge) error {
	if charge.CreatedAt.IsZero() {
		charge.CreatedAt = time.Now()
	}

	charge.UpdatedAt = time.Now()

	for _, i := range charge.Items {
		if i.ID.IsZero() {
			i.ID = primitive.NewObjectID()
		}
	}

	_, err := repo.collectionCharges.ReplaceOne(
		ctx,
		bson.M{"_id": charge.ID},
		charge,
		options.Replace().SetUpsert(true),
	)

	return err
}

func (repo *repositoryImpl) ChargeRemove(ctx context.Context, id string) error {
	_, err := repo.collectionCharges.DeleteOne(ctx, bson.M{"_id": id})

	return err
}

func (repo *repositoryImpl) ChargeGetByID(ctx context.Context, id string) (*model.Charge, error) {
	m := &model.Charge{}

	err := repo.collectionCharges.FindOne(ctx, bson.M{"_id": id}).Decode(m)
	if err != nil {
		return nil, err
	}

	return m, nil
}

func (repo *repositoryImpl) ChargeGetBySessionID(ctx context.Context, id string) (*model.Charge, error) {
	m := &model.Charge{}

	err := repo.collectionCharges.FindOne(ctx, bson.M{"session_id": id}).Decode(m)
	if err != nil {
		return nil, err
	}

	return m, nil
}

func (repo *repositoryImpl) ChargeGetBySubscriptionID(ctx context.Context, id string) (*model.Charge, error) {
	m := &model.Charge{}

	err := repo.collectionCharges.FindOne(ctx, bson.M{"subscription_id": id}).Decode(m)
	if err != nil {
		return nil, err
	}

	return m, nil
}

func (repo *repositoryImpl) MandateSave(ctx context.Context, mandate *model.Mandate) error {
	if mandate.ID.IsZero() {
		mandate.ID = primitive.NewObjectID()
	}

	if mandate.CreatedAt.IsZero() {
		mandate.CreatedAt = time.Now()
	}

	mandate.UpdatedAt = time.Now()

	_, err := repo.collectionMandates.ReplaceOne(
		ctx,
		bson.M{"_id": mandate.ID},
		mandate,
		options.Replace().SetUpsert(true),
	)

	return err
}

func (repo *repositoryImpl) MandateRemove(ctx context.Context, id string) error {
	_, err := repo.collectionMandates.DeleteOne(ctx, bson.M{"_id": id})

	return err
}

func (repo *repositoryImpl) MandateGetByID(ctx context.Context, id primitive.ObjectID) (*model.Mandate, error) {
	m := &model.Mandate{}

	err := repo.collectionMandates.FindOne(ctx, bson.M{"_id": id}).Decode(m)
	if err != nil {
		return nil, err
	}

	return m, nil
}

func (repo *repositoryImpl) MandateGetBySetupIntentID(
	ctx context.Context,
	setupIntentID string,
) (*model.Mandate, error) {
	m := &model.Mandate{}

	err := repo.collectionMandates.FindOne(ctx, bson.M{"setup_intent_id": setupIntentID}).Decode(m)
	if err != nil {
		return nil, err
	}

	return m, nil
}

func (repo *repositoryImpl) MandateGetByMandateID(ctx context.Context, mandateID string) (*model.Mandate, error) {
	m := &model.Mandate{}

	err := repo.collectionMandates.FindOne(ctx, bson.M{"mandate_id": mandateID}).Decode(m)
	if err != nil {
		return nil, err
	}

	return m, nil
}

func (repo *repositoryImpl) MandateGetByUserIDEmail(
	ctx context.Context,
	userID, email string,
) (m *model.Mandate, err error) {
	m = &model.Mandate{}

	err = repo.collectionMandates.FindOne(ctx, bson.M{"user_id": userID, "email": email}).Decode(m)
	if err != nil {
		return nil, err
	}

	return
}

func (repo *repositoryImpl) MandateGetByCustomerID(
	ctx context.Context,
	customerID string,
) (m *model.Mandate, err error) {
	m = &model.Mandate{}

	err = repo.collectionMandates.FindOne(ctx, bson.M{"customer_id": customerID}).Decode(m)
	if err != nil {
		return nil, err
	}

	return
}

func (repo *repositoryImpl) FilterMandates(
	ctx context.Context,
	params *query.QueryParams,
) (models []*model.Mandate, count int, err error) {
	filter, opts, err := mongoquery.BuildFilter(model.Mandate{}, params, nil)
	if err != nil {
		return nil, 0, err
	}

	cur, err := repo.collectionMandates.Find(ctx, filter, opts)
	if err != nil {
		return nil, 0, err
	}

	err = cur.All(ctx, &models)
	if err != nil {
		return nil, 0, err
	}

	total, err := repo.collectionMandates.CountDocuments(ctx, filter)
	if err != nil {
		return nil, 0, err
	}

	count = int(total)

	return
}

func (repo *repositoryImpl) UserOffersSave(ctx context.Context, u *model.UserOffers) error {
	_, err := repo.collectionUserOffers.ReplaceOne(
		ctx,
		bson.M{"offer": u.Offer},
		u,
		options.Replace().SetUpsert(true),
	)

	return err
}

func (repo *repositoryImpl) UserOffersGet(ctx context.Context, offer string) (*model.UserOffers, error) {
	m := &model.UserOffers{}

	err := repo.collectionUserOffers.FindOne(ctx, bson.M{"offer": offer}).Decode(m)
	if err != nil {
		return nil, err
	}

	return m, nil
}

func (repo *repositoryImpl) UserOffersGetByUser(ctx context.Context,
	user string) (models []*model.UserOffers, count int, err error) {
	filter := bson.M{"users.id": user}

	cur, err := repo.collectionUserOffers.Find(ctx, filter)
	if err != nil {
		return nil, 0, err
	}

	err = cur.All(ctx, &models)
	if err != nil {
		return nil, 0, err
	}

	total, err := repo.collectionUserOffers.CountDocuments(ctx, filter)
	if err != nil {
		return nil, 0, err
	}

	count = int(total)

	return
}

func (repo *repositoryImpl) FilterUserOffers(
	ctx context.Context,
	params *query.QueryParams,
) (models []*model.UserOffers, count int, err error) {
	filter, opts, err := mongoquery.BuildFilter(model.UserOffers{}, params, nil)
	if err != nil {
		return nil, 0, err
	}

	cur, err := repo.collectionUserOffers.Find(ctx, filter, opts)
	if err != nil {
		return nil, 0, err
	}

	err = cur.All(ctx, &models)
	if err != nil {
		return nil, 0, err
	}

	total, err := repo.collectionUserOffers.CountDocuments(ctx, filter)
	if err != nil {
		return nil, 0, err
	}

	count = int(total)

	return
}

func (repo *repositoryImpl) FilterCharges(
	ctx context.Context,
	params *query.QueryParams,
) (models []*model.Charge, count int, err error) {
	filter, opts, err := mongoquery.BuildFilter(model.Charge{}, params, nil)
	if err != nil {
		return nil, 0, err
	}

	cur, err := repo.collectionCharges.Find(ctx, filter, opts)
	if err != nil {
		return nil, 0, err
	}

	err = cur.All(ctx, &models)
	if err != nil {
		return nil, 0, err
	}

	total, err := repo.collectionCharges.CountDocuments(ctx, filter)
	if err != nil {
		return nil, 0, err
	}

	count = int(total)

	return
}

func (repo *repositoryImpl) ChangelogSave(ctx context.Context, changelog *model.Changelog) error {
	changelog.Created = time.Now()

	res, err := repo.collectionChangelog.InsertOne(ctx, changelog)
	if err != nil {
		return err
	}

	changelog.ID, _ = res.InsertedID.(primitive.ObjectID)

	return err
}
