// Package service provides implementation for cardpayment service
package service

import (
	"context"
	"fmt"

	"gitlab.mashplace.com/mashplace/backend/common/api/grpcctx"
	"gitlab.mashplace.com/mashplace/backend/common/generic/log"

	"github.com/camunda-cloud/zeebe/clients/go/pkg/zbc"
	cardpayment "gitlab.mashplace.com/mashplace/backend/cardpayment/api"
	"gitlab.mashplace.com/mashplace/backend/cardpayment/internal/model"
)

// Stripe provides configration for stripe client
type Stripe struct {
	Currency           stripe.Currency
	InsuranceAccount   string
	CallbackSuccessURL string
	CallbackCancelURL  string
	WebhookSecret      string
}

// Config for service implementation
type Config struct {
	// Persistent repository
	Repository model.Repository
	// ZeebeClient BPM client
	ZeebeClient zbc.Client
	// Stripe settings
	Stripe Stripe
}

// Service implementation
type Service interface {
	cardpayment.CardpaymentServiceServer
	RPCCallback(ctx context.Context, rpc, op, svc, reqid string, values, metadata map[string]interface{})
}

func validateConfigStripe(stripe *Stripe) error {
	if stripe.InsuranceAccount == "" {
		return fmt.Errorf("no currency in stripe config")
	}

	if stripe.InsuranceAccount == "" {
		return fmt.Errorf("no insurance account in stripe config")
	}

	if stripe.CallbackSuccessURL == "" {
		return fmt.Errorf("no callback success URL in stripe config")
	}

	if stripe.CallbackCancelURL == "" {
		return fmt.Errorf("no callback cancel URL in stripe config")
	}

	if stripe.WebhookSecret == "" {
		return fmt.Errorf("no webhook secret in stripe config")
	}

	return nil
}

func validateConfig(config *Config) error {
	if config == nil {
		return fmt.Errorf("invalid service config")
	}

	if config.Repository == nil {
		return fmt.Errorf("no repository in service config")
	}

	err := validateConfigStripe(&config.Stripe)
	if err != nil {
		return err
	}

	return nil
}

// NewService returns new service instance
func NewService(ctx context.Context, config *Config) (Service, error) {
	err := validateConfig(config)
	if err != nil {
		return nil, err
	}

	srv := &serviceImpl{
		config:        *config,
		notifyClients: make(map[string]cardpayment.CardpaymentNotifyServiceClient),
	}

	return srv, nil
}

func (srv *serviceImpl) RPCCallback(
	ctx context.Context,
	rpc, op, svc, reqid string,
	values, metadata map[string]interface{},
) {
	trace := grpcctx.ContextTraceID(ctx)

	userid := grpcctx.ContextUserID(ctx)

	if userid == "" {
		userid = grpcctx.ContextSessionID(ctx)
	}

	if metadata == nil {
		metadata = make(map[string]interface{})
	}

	metadata["user"] = userid

	err := srv.config.Repository.ChangelogSave(ctx, &model.Changelog{
		Context:  trace,
		Kind:     rpc,
		Object:   svc,
		Ref:      reqid,
		Op:       op,
		Old:      nil,
		New:      values,
		Metadata: metadata,
	})
	if err != nil {
		log.WithError(err).Error("saving rpc changelog")
	}
}
