package service

import (
	"bytes"
	"context"
	"crypto/sha256"
	"encoding/json"
	"errors"
	"io"
	"strings"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/api/grpcctx"
	"gitlab.mashplace.com/mashplace/backend/common/generic/grpc"

	"github.com/stripe/stripe-go/v72/sub"

	"github.com/stripe/stripe-go/v72/customer"

	"github.com/stripe/stripe-go/v72/paymentmethod"
	"gitlab.mashplace.com/mashplace/backend/common/api/generic/query"
	"gitlab.mashplace.com/mashplace/backend/common/generic/log"

	"go.mongodb.org/mongo-driver/mongo"

	"github.com/stripe/stripe-go/v72/mandate"
	"github.com/stripe/stripe-go/v72/paymentintent"
	"github.com/stripe/stripe-go/v72/setupintent"
	"github.com/stripe/stripe-go/v72/webhook"

	cardpayment "gitlab.mashplace.com/mashplace/backend/cardpayment/api"
	"gitlab.mashplace.com/mashplace/backend/cardpayment/internal/model"
	"gitlab.mashplace.com/mashplace/backend/common/api/grpcerr"
)

var errInvalidInit = grpcerr.New(grpcerr.ProviderError, "invalid init")
var errInvalidChunk = grpcerr.New(grpcerr.ProviderError, "invalid chunk")
var errInvalidHash = grpcerr.New(grpcerr.ProviderError, "invalid hash")

type stripeCallback struct {
	Created         int    `json:"created"`
	Livemode        bool   `json:"livemode"`
	ID              string `json:"id"`
	Type            string `json:"type"`
	Object          string `json:"object"`
	PendingWebhooks int    `json:"pending_webhooks"`
	APIVersion      string `json:"api_version"`
}

type webhookResponse struct {
	ID     string
	Status cardpayment.PaymentStatus
}

type stripeCharge struct {
	Created         int         `json:"created"`
	Livemode        bool        `json:"livemode"`
	ID              string      `json:"id"`
	Type            string      `json:"type"`
	Object          string      `json:"object"`
	Request         interface{} `json:"request"`
	PendingWebhooks int         `json:"pending_webhooks"`
	APIVersion      string      `json:"api_version"`
	Data            struct {
		Object struct {
			FraudDetails struct {
			} `json:"fraud_details"`
			Metadata struct {
			} `json:"metadata"`
			Refunds struct {
				Object     string        `json:"object"`
				Data       []interface{} `json:"data"`
				HasMore    bool          `json:"has_more"`
				TotalCount int           `json:"total_count"`
				URL        string        `json:"url"`
			} `json:"refunds"`
			Source struct {
				ID                string  `json:"id"`
				Object            string  `json:"object"`
				AddressCity       *string `json:"address_city"`
				AddressCountry    *string `json:"address_country"`
				AddressLine1      *string `json:"address_line1"`
				AddressLine1Check *string `json:"address_line1_check"`
				AddressLine2      *string `json:"address_line2"`
				AddressState      *string `json:"address_state"`
				AddressZip        string  `json:"address_zip"`
				AddressZipCheck   string  `json:"address_zip_check"`
				Brand             string  `json:"brand"`
				Country           string  `json:"country"`
				Customer          *string `json:"customer"`
				CvcCheck          string  `json:"cvc_check"`
				DynamicLast4      *string `json:"dynamic_last4"`
				ExpMonth          int     `json:"exp_month"`
				ExpYear           int     `json:"exp_year"`
				Fingerprint       string  `json:"fingerprint"`
				Funding           string  `json:"funding"`
				Last4             string  `json:"last4"`
				Metadata          struct {
				} `json:"metadata"`
				Name               string  `json:"name"`
				TokenizationMethod *string `json:"tokenization_method"`
			} `json:"source"`
			ID                   string      `json:"id"`
			Object               string      `json:"object"`
			BalanceTransaction   string      `json:"balance_transaction"`
			Currency             string      `json:"currency"`
			Description          string      `json:"description"`
			Status               string      `json:"status"`
			Shipping             interface{} `json:"shipping"`
			Outcome              interface{} `json:"outcome"`
			TransferData         interface{} `json:"transfer_data"`
			Amount               int         `json:"amount"`
			AmountRefunded       int         `json:"amount_refunded"`
			Created              int         `json:"created"`
			Application          *string     `json:"application"`
			ApplicationFee       *string     `json:"application_fee"`
			ApplicationFeeAmount *int64      `json:"application_fee_amount"`
			Customer             *string     `json:"customer"`
			Destination          *string     `json:"destination"`
			Dispute              *string     `json:"dispute"`
			FailureCode          *string     `json:"failure_code"`
			FailureMessage       *string     `json:"failure_message"`
			Invoice              *string     `json:"invoice"`
			OnBehalfOf           *string     `json:"on_behalf_of"`
			Order                *string     `json:"order"`
			PaymentIntent        *string     `json:"payment_intent"`
			ReceiptEmail         *string     `json:"receipt_email"`
			ReceiptNumber        *string     `json:"receipt_number"`
			Review               *string     `json:"review"`
			SourceTransfer       *string     `json:"source_transfer"`
			StatementDescriptor  *string     `json:"statement_descriptor"`
			TransferGroup        *string     `json:"transfer_group"`
			Livemode             bool        `json:"livemode"`
			Paid                 bool        `json:"paid"`
			Refunded             bool        `json:"refunded"`
			Captured             bool        `json:"captured"`
		} `json:"object"`
	} `json:"data"`
}

type stripeSource struct {
	Created         int         `json:"created"`
	Livemode        bool        `json:"livemode"`
	ID              string      `json:"id"`
	Type            string      `json:"type"`
	Object          string      `json:"object"`
	Request         interface{} `json:"request"`
	PendingWebhooks int         `json:"pending_webhooks"`
	APIVersion      string      `json:"api_version"`
	Data            struct {
		Object struct {
			ID                string `json:"id"`
			Object            string `json:"object"`
			AchCreditTransfer struct {
				AccountNumber string `json:"account_number"`
				RoutingNumber string `json:"routing_number"`
				Fingerprint   string `json:"fingerprint"`
				BankName      string `json:"bank_name"`
				SwiftCode     string `json:"swift_code"`
			} `json:"ach_credit_transfer"`
			Amount       interface{} `json:"amount"`
			ClientSecret string      `json:"client_secret"`
			Created      int         `json:"created"`
			Currency     string      `json:"currency"`
			Flow         string      `json:"flow"`
			Livemode     bool        `json:"livemode"`
			Metadata     struct {
			} `json:"metadata"`
			Owner struct {
				Address         interface{} `json:"address"`
				Email           string      `json:"email"`
				Name            interface{} `json:"name"`
				Phone           interface{} `json:"phone"`
				VerifiedAddress interface{} `json:"verified_address"`
				VerifiedEmail   interface{} `json:"verified_email"`
				VerifiedName    interface{} `json:"verified_name"`
				VerifiedPhone   interface{} `json:"verified_phone"`
			} `json:"owner"`
			Receiver struct {
				Address                string `json:"address"`
				AmountCharged          int    `json:"amount_charged"`
				AmountReceived         int    `json:"amount_received"`
				AmountReturned         int    `json:"amount_returned"`
				RefundAttributesMethod string `json:"refund_attributes_method"`
				RefundAttributesStatus string `json:"refund_attributes_status"`
			} `json:"receiver"`
			StatementDescriptor interface{} `json:"statement_descriptor"`
			Status              string      `json:"status"`
			Type                string      `json:"type"`
			Usage               string      `json:"usage"`
		} `json:"object"`
	} `json:"data"`
}

type paymentIntent struct {
	Created         int         `json:"created"`
	Livemode        bool        `json:"livemode"`
	ID              string      `json:"id"`
	Type            string      `json:"type"`
	Object          string      `json:"object"`
	Request         interface{} `json:"request"`
	PendingWebhooks int         `json:"pending_webhooks"`
	APIVersion      string      `json:"api_version"`
	Data            struct {
		Object struct {
			ID                   string      `json:"id"`
			Object               string      `json:"object"`
			AllowedSourceTypes   []string    `json:"allowed_source_types"`
			Amount               int         `json:"amount"`
			AmountCapturable     int         `json:"amount_capturable"`
			AmountReceived       int         `json:"amount_received"`
			Application          interface{} `json:"application"`
			ApplicationFeeAmount interface{} `json:"application_fee_amount"`
			CanceledAt           interface{} `json:"canceled_at"`
			CancellationReason   interface{} `json:"cancellation_reason"`
			CaptureMethod        string      `json:"capture_method"`
			Charges              struct {
				Object     string        `json:"object"`
				Data       []interface{} `json:"data"`
				HasMore    bool          `json:"has_more"`
				TotalCount int           `json:"total_count"`
				URL        string        `json:"url"`
			} `json:"charges"`
			ClientSecret       interface{} `json:"client_secret"`
			ConfirmationMethod string      `json:"confirmation_method"`
			Created            int         `json:"created"`
			Currency           string      `json:"currency"`
			Customer           interface{} `json:"customer"`
			Description        string      `json:"description"`
			LastPaymentError   interface{} `json:"last_payment_error"`
			Livemode           bool        `json:"livemode"`
			Metadata           struct {
				OrderID string `json:"order_id"`
			} `json:"metadata"`
			NextSourceAction interface{} `json:"next_source_action"`
			OnBehalfOf       interface{} `json:"on_behalf_of"`
			ReceiptEmail     string      `json:"receipt_email"`
			Review           interface{} `json:"review"`
			Shipping         struct {
				Address struct {
					City       string      `json:"city"`
					Country    string      `json:"country"`
					Line1      string      `json:"line1"`
					Line2      interface{} `json:"line2"`
					PostalCode string      `json:"postal_code"`
					State      interface{} `json:"state"`
				} `json:"address"`
				Carrier        interface{} `json:"carrier"`
				Name           string      `json:"name"`
				Phone          interface{} `json:"phone"`
				TrackingNumber interface{} `json:"tracking_number"`
			} `json:"shipping"`
			Source              string      `json:"source"`
			StatementDescriptor string      `json:"statement_descriptor"`
			Status              string      `json:"status"`
			TransferData        interface{} `json:"transfer_data"`
			TransferGroup       interface{} `json:"transfer_group"`
		} `json:"object"`
	} `json:"data"`
}

func getStatus(customStatus string) cardpayment.PaymentStatus {
	if strings.Contains(customStatus, "succeeded") || strings.Contains(customStatus, "success") {
		return cardpayment.PaymentStatus_SUCCEEDED
	}

	if strings.Contains(customStatus, "failed") {
		return cardpayment.PaymentStatus_FAILED
	}

	return cardpayment.PaymentStatus_UNKNOWN_PAYMENT_STATUS
}

func (srv *serviceImpl) handleCallback(ctx context.Context, bytes []byte) (result *webhookResponse, err error) {
	var vals map[string]interface{}

	_ = json.Unmarshal(bytes, &vals)

	trace := grpcctx.ContextTraceID(ctx)

	reqid := grpc.MethodRequestId(ctx)

	err = srv.config.Repository.ChangelogSave(ctx, &model.Changelog{
		Context:  trace,
		Kind:     "callback",
		Object:   "stripe",
		Ref:      reqid,
		Op:       "call",
		Old:      nil,
		New:      vals,
		Metadata: nil,
	})
	if err != nil {
		log.WithError(err).Error("saving changelog")
	}

	var cb stripeCallback

	err = json.Unmarshal(bytes, &cb)
	if err != nil {
		return nil, err
	}

	result = &webhookResponse{}

	switch {
	case strings.HasPrefix(cb.Type, "charge"):
		var ch stripeCharge

		err = json.Unmarshal(bytes, &ch)
		if err != nil {
			return nil, err
		}

		result.ID = ch.Data.Object.ID
		result.Status = getStatus(ch.Data.Object.Status)
	case strings.HasPrefix(cb.Type, "source"):
		var src stripeSource

		err = json.Unmarshal(bytes, &src)
		if err != nil {
			return nil, err
		}

		result.ID = src.Data.Object.ID
		result.Status = getStatus(src.Data.Object.Status)
	case strings.HasPrefix(cb.Type, "payment_intent"):
		var pi paymentIntent

		err = json.Unmarshal(bytes, &pi)
		if err != nil {
			return nil, err
		}

		result.ID = pi.Data.Object.ID
		result.Status = getStatus(pi.Data.Object.Status)
	default:
		return nil, grpcerr.Newf(grpcerr.InvalidData, "unknown callback type: '%s'", cb.Type)
	}

	return
}

func (srv *serviceImpl) HandleCallback(stream cardpayment.CardpaymentService_HandleCallbackServer) error {
	ev, err := stream.Recv()
	if err != nil {
		return err
	}

	if ev.Init == nil {
		return errInvalidInit
	}

	remote := ev.Init.Sha256

	var data []byte

	for {
		ev, err = stream.Recv()
		if err != nil {
			if errors.Is(err, io.EOF) || errors.Is(err, io.ErrUnexpectedEOF) {
				break
			}

			return err
		}

		if ev.Chunk == nil {
			return errInvalidChunk
		}

		if len(ev.Chunk.Data) == 0 {
			break
		}

		data = append(data, ev.Chunk.Data...)
	}

	h := sha256.New()
	actual := h.Sum(data)

	if !bytes.Equal(remote, actual) {
		return errInvalidHash
	}

	resp, err := srv.handleCallback(stream.Context(), data)
	if err != nil {
		return err
	}

	m, err := srv.config.Repository.ChargeGetByID(stream.Context(), resp.ID)
	if err != nil {
		return err
	}

	m.Status = resp.Status

	err = srv.config.Repository.ChargeSave(stream.Context(), m)
	if err != nil {
		return err
	}

	go srv.notifyUpdate(context.Background(), m)

	return nil
}

func (srv *serviceImpl) handleWebhookPaymentIntentStatus(
	ctx context.Context,
	event *stripe.Event,
	status cardpayment.PaymentStatus,
) (err error) {
	var p stripe.PaymentIntent

	err = json.Unmarshal(event.Data.Raw, &p)
	if err != nil {
		return err
	}

	log.Infof("payment_intent got PaymentIntent object : %+v\n", p)

	var intentModel *model.Charge

	intentModel, err = srv.config.Repository.ChargeGetByID(ctx, p.ID)
	if err != nil {
		return err
	}

	if intentModel.Status == cardpayment.PaymentStatus_SUCCEEDED {
		go srv.notifyUpdate(context.Background(), intentModel)

		return nil
	}

	intentModel.Status = status

	err = srv.config.Repository.ChargeSave(ctx, intentModel)
	if err != nil {
		return err
	}

	go srv.notifyUpdate(context.Background(), intentModel)

	return nil
}

func (srv *serviceImpl) handleWebhookChargeStatus(
	ctx context.Context,
	event *stripe.Event,
	status cardpayment.PaymentStatus,
) (err error) {
	var ch stripe.Charge

	err = json.Unmarshal(event.Data.Raw, &ch)
	if err != nil {
		return err
	}

	var p *model.Charge

	switch {
	case ch.PaymentIntent != nil:
		p, err = srv.config.Repository.ChargeGetByID(ctx, ch.PaymentIntent.ID)
		if err != nil {
			return err
		}
	default:
		p, err = srv.config.Repository.ChargeGetByID(ctx, ch.ID)
		if err != nil {
			return err
		}
	}

	if p.Status == cardpayment.PaymentStatus_SUCCEEDED {
		return nil
	}

	p.Status = status

	err = srv.config.Repository.ChargeSave(ctx, p)
	if err != nil {
		return err
	}

	return nil
}

func (srv *serviceImpl) handleWebhookPaymentMethodAutoUpdate(ctx context.Context, event *stripe.Event) (err error) {
	var sessionObj stripe.CheckoutSession

	err = json.Unmarshal(event.Data.Raw, &sessionObj)
	if err != nil {
		return err
	}

	var m *model.Charge

	switch {
	case sessionObj.PaymentIntent != nil:
		m, err = srv.config.Repository.ChargeGetByID(ctx, sessionObj.PaymentIntent.ID)
		if err != nil {
			return err
		}
	default:
		m, err = srv.config.Repository.ChargeGetByID(ctx, sessionObj.ID)
		if err != nil {
			return err
		}
	}

	m.Status = cardpayment.PaymentStatus_FAILED

	err = srv.config.Repository.ChargeSave(ctx, m)
	if err != nil {
		return err
	}

	return nil
}

func (srv *serviceImpl) handleWebhookMandateUpdateResolveMandateCustomerID(
	ctx context.Context,
	m *stripe.Mandate,
) (mandateModel *model.Mandate, customerID string, err error) {
	mandateModel, err = srv.config.Repository.MandateGetByMandateID(ctx, m.ID)

	var pm *stripe.PaymentMethod

	// customer info may not be available for several seconds after receiving mandate.update
	// but do not retry indefinitely
	for i := 0; i < 300 && mandateModel == nil; i++ {
		pm, err = paymentmethod.Get(m.PaymentMethod.ID, nil)
		if err != nil {
			log.WithError(err).Errorf("mandate.updated can't found paymentmethod object by ID : %v\n", m.PaymentMethod.ID)

			return
		}

		log.Infof("mandate.updated found paymentmethod (%s) object with customer: %+v\n", m.PaymentMethod.ID, pm.Customer)

		if pm.Customer != nil && pm.Customer.ID != "" {
			customerID = pm.Customer.ID

			mandateModel, err = srv.config.Repository.MandateGetByCustomerID(ctx, pm.Customer.ID)

			break
		}

		time.Sleep(time.Second)
	}

	if errors.Is(err, mongo.ErrNoDocuments) {
		err = nil
	}

	if err != nil {
		log.WithError(err).Errorf("mandate.updated not found mandate model by ID : %v\n", m.ID)
		return
	}

	log.Infof("mandate.updated found mandate model : %v\n", mandateModel)

	if mandateModel == nil && customerID == "" {
		log.WithError(err).Errorln("mandate.updated not found mandate model and customer ID")

		return nil, "", errEmptyCustomerID
	}

	switch {
	case mandateModel == nil && customerID == "":
		log.WithError(err).Errorln("mandate.updated not found mandate model and customer ID")

		return nil, "", errEmptyCustomerID
	case mandateModel == nil:
		mandateModel = &model.Mandate{
			MandateID:       m.ID,
			PaymentMethodID: m.PaymentMethod.ID,
			CustomerID:      customerID,
			Valid:           false,
			Status:          cardpayment.MandateStatus_PENDING,
			CreatedAt:       time.Now(),
			UpdatedAt:       time.Now(),
		}
	case customerID == "":
		customerID = mandateModel.CustomerID
	}

	return
}

func (srv *serviceImpl) handleWebhookMandateUpdateResolveUser(
	customerID string,
	mandateModel *model.Mandate,
) (err error) {
	if (mandateModel.UserID != "" && mandateModel.Email != "") || customerID == "" {
		return
	}

	cus, err := customer.Get(customerID, nil)
	if err != nil {
		log.WithError(err).Errorf("mandate.updated can't found customer object by ID : %v\n", customerID)

		return
	}

	if mandateModel.UserID == "" {
		mandateModel.UserID = cus.Description
	}

	if mandateModel.Email == "" {
		mandateModel.Email = cus.Email
	}

	return
}

func (srv *serviceImpl) handleWebhookMandateUpdateResolveOffer(
	ctx context.Context,
	mandateModel *model.Mandate,
) (err error) {
	if (len(mandateModel.Offers) > 0 && mandateModel.Offers[0] != "") || mandateModel.UserID == "" {
		return
	}

	offs, cnt, err := srv.config.Repository.UserOffersGetByUser(ctx, mandateModel.UserID)
	if err != nil {
		log.WithError(err).Errorf("mandate.updated not found offers by user ID : %v\n", mandateModel.UserID)

		return err
	}

	var offers []string

	if cnt > 0 {
		for _, o := range offs {
			if o.Offer != "" {
				offers = append(offers, o.Offer)
			}
		}
	}

	mandateModel.Offers = offers

	return
}

func (srv *serviceImpl) handleWebhookMandateUpdated(ctx context.Context, event *stripe.Event) (err error) {
	var m stripe.Mandate

	err = json.Unmarshal(event.Data.Raw, &m)
	if err != nil {
		return err
	}

	log.Infof("mandate.updated got mandate object : %+v\n", m)

	mandateModel, customerID, err := srv.handleWebhookMandateUpdateResolveMandateCustomerID(ctx, &m)
	if err != nil {
		return err
	}

	if mandateModel.MandateID == "" {
		mandateModel.MandateID = m.ID
	}

	if mandateModel.PaymentMethodID == "" {
		mandateModel.PaymentMethodID = m.PaymentMethod.ID
	}

	if mandateModel.CustomerID == "" {
		mandateModel.CustomerID = customerID
	}

	err = srv.handleWebhookMandateUpdateResolveUser(customerID, mandateModel)
	if err != nil {
		return err
	}

	err = srv.handleWebhookMandateUpdateResolveOffer(ctx, mandateModel)
	if err != nil {
		return err
	}

	log.Infof("mandate.updated got status : %v for mandate : %v\n", m.Status, mandateModel.MandateID)

	if m.Status == stripe.MandateStatusActive {
		mandateModel.Status = cardpayment.MandateStatus_ACTIVE
		mandateModel.Valid = true
	} else if m.Status == stripe.MandateStatusInactive {
		mandateModel.Status = cardpayment.MandateStatus_INACTIVE
		mandateModel.Valid = false
	}

	err = srv.config.Repository.MandateSave(ctx, mandateModel)
	if err != nil {
		log.WithError(err).Errorf("mandate.updated error on save mandate model : %v\n", mandateModel)
		return err
	}

	log.Infof("mandate.updated save mandate model : %v\n", mandateModel)

	return nil
}

func (srv *serviceImpl) handleWebhookCheckoutSessionFailed(ctx context.Context, event *stripe.Event) (err error) {
	var sessionObj stripe.CheckoutSession

	err = json.Unmarshal(event.Data.Raw, &sessionObj)
	if err != nil {
		return err
	}

	log.Infof("checkout.session.async_payment_failed got session object : %+v\n", sessionObj)

	var m *model.Charge

	switch {
	case sessionObj.PaymentIntent != nil:
		m, err = srv.config.Repository.ChargeGetByID(ctx, sessionObj.PaymentIntent.ID)
		if err != nil {
			return err
		}
	default:
		m, err = srv.config.Repository.ChargeGetByID(ctx, sessionObj.ID)
		if err != nil {
			return err
		}
	}

	m.Status = cardpayment.PaymentStatus_FAILED

	err = srv.config.Repository.ChargeSave(ctx, m)
	if err != nil {
		return err
	}

	return nil
}

func (srv *serviceImpl) handleWebhookCheckoutSessionSucceeded(ctx context.Context, event *stripe.Event) (err error) {
	var sessionObj stripe.CheckoutSession

	err = json.Unmarshal(event.Data.Raw, &sessionObj)
	if err != nil {
		return err
	}

	log.Infof("checkout.session.async_payment_succeeded got session object : %+v\n", sessionObj)

	var m *model.Charge

	switch {
	case sessionObj.PaymentIntent != nil:
		m, err = srv.config.Repository.ChargeGetByID(ctx, sessionObj.PaymentIntent.ID)
		if err != nil {
			return err
		}
	case sessionObj.Subscription != nil:
		m, err = srv.config.Repository.ChargeGetByID(ctx, sessionObj.ID)
		if err != nil {
			return err
		}

		m.SubscriptionID = sessionObj.Subscription.ID
	default:
		m, err = srv.config.Repository.ChargeGetByID(ctx, sessionObj.ID)
		if err != nil {
			return err
		}
	}

	m.Status = cardpayment.PaymentStatus_SUCCEEDED

	err = srv.config.Repository.ChargeSave(ctx, m)
	if err != nil {
		return err
	}

	go srv.notifyUpdate(context.Background(), m)

	return nil
}

func (srv *serviceImpl) handleWebhookCheckoutSessionCompletedSetupIntent(
	ctx context.Context,
	sessionObj *stripe.CheckoutSession,
) (err error) {
	var sintent *stripe.SetupIntent

	sintent, err = setupintent.Get(sessionObj.SetupIntent.ID, nil)
	if err != nil || sintent == nil || sintent.Mandate == nil {
		return err
	}

	var mandateModel *model.Mandate

	mandateModel, err = srv.config.Repository.MandateGetBySetupIntentID(ctx, sintent.ID)
	if err != nil {
		return err
	}

	mandateModel.MandateID = sintent.Mandate.ID

	var mandateObj *stripe.Mandate

	mandateObj, err = mandate.Get(sintent.Mandate.ID, nil)
	if err != nil {
		return err
	}

	if mandateModel.PaymentMethodID == "" && mandateObj.PaymentMethod != nil {
		mandateModel.PaymentMethodID = mandateObj.PaymentMethod.ID
	}

	switch mandateObj.Status {
	case stripe.MandateStatusActive:
		mandateModel.Status = cardpayment.MandateStatus_ACTIVE
		mandateModel.Valid = true
	case stripe.MandateStatusInactive:
		mandateModel.Status = cardpayment.MandateStatus_INACTIVE
		mandateModel.Valid = false
	case stripe.MandateStatusPending:
		mandateModel.Status = cardpayment.MandateStatus_PENDING
	}

	err = srv.config.Repository.MandateSave(ctx, mandateModel)
	if err != nil {
		return err
	}

	return nil
}

func (srv *serviceImpl) handleWebhookCheckoutSessionCompletedPaymentIntent(
	ctx context.Context,
	sessionObj *stripe.CheckoutSession,
) (err error) {
	var intent *stripe.PaymentIntent

	intent, err = paymentintent.Get(sessionObj.PaymentIntent.ID, nil)
	if err != nil || intent == nil {
		return err
	}

	if intent.Status == stripe.PaymentIntentStatusSucceeded {
		var m *model.Charge

		m, err = srv.config.Repository.ChargeGetByID(ctx, intent.ID)
		if err != nil {
			return err
		}

		m.Status = cardpayment.PaymentStatus_SUCCEEDED

		err = srv.config.Repository.ChargeSave(ctx, m)
		if err != nil {
			return err
		}
	}

	return nil
}

func (srv *serviceImpl) handleWebhookCheckoutSessionCompletedSubscription(
	ctx context.Context,
	sessionObj *stripe.CheckoutSession,
) (err error) {
	var subscription *stripe.Subscription

	subscription, err = sub.Get(sessionObj.Subscription.ID, nil)
	if err != nil || subscription == nil {
		return err
	}

	if subscription.Status == stripe.SubscriptionStatusActive {
		var m *model.Charge

		m, err = srv.config.Repository.ChargeGetByID(ctx, sessionObj.ID)
		if err != nil {
			return err
		}

		m.Status = cardpayment.PaymentStatus_SUCCEEDED
		m.SubscriptionID = sessionObj.Subscription.ID

		err = srv.config.Repository.ChargeSave(ctx, m)
		if err != nil {
			return err
		}

		go srv.notifyUpdate(context.Background(), m)
	}

	return nil
}

func (srv *serviceImpl) handleWebhookCheckoutSessionCompleted(ctx context.Context, event *stripe.Event) (err error) {
	var sessionObj stripe.CheckoutSession

	err = json.Unmarshal(event.Data.Raw, &sessionObj)
	if err != nil {
		return err
	}

	switch {
	case sessionObj.SetupIntent != nil:
		err = srv.handleWebhookCheckoutSessionCompletedSetupIntent(ctx, &sessionObj)
		if err != nil {
			return err
		}
	case sessionObj.PaymentIntent != nil:
		err = srv.handleWebhookCheckoutSessionCompletedPaymentIntent(ctx, &sessionObj)
		if err != nil {
			return err
		}
	case sessionObj.Subscription != nil:
		err = srv.handleWebhookCheckoutSessionCompletedSubscription(ctx, &sessionObj)
		if err != nil {
			return err
		}
	}

	return nil
}

func (srv *serviceImpl) HandleWebhook(
	ctx context.Context,
	req *cardpayment.HandleWebhookRequest,
) (resp *cardpayment.HandleWebhookResponse, err error) {
	defer func() {
		if errors.Is(err, mongo.ErrNoDocuments) {
			log.WithError(err).Debugf("HandleWebhook suppressing mongo.ErrNoDocuments\n")

			err = nil
			resp = &cardpayment.HandleWebhookResponse{
				Success: true,
				Status:  cardpayment.HttpStatusCode_OK,
			}
		}
	}()

	event, err := webhook.ConstructEvent(req.GetData(), req.GetStripeSignature(), srv.config.Stripe.WebhookSecret)
	if err != nil {
		return &cardpayment.HandleWebhookResponse{
			Success: false,
			Status:  cardpayment.HttpStatusCode_BAD_REQUEST,
		}, err
	}

	log.Infof("HandleWebhook got event %v\n", event.Type)

	var vals map[string]interface{}

	_ = json.Unmarshal(req.GetData(), &vals)

	trace := grpcctx.ContextTraceID(ctx)

	reqid := grpc.MethodRequestId(ctx)

	err = srv.config.Repository.ChangelogSave(ctx, &model.Changelog{
		Context:  trace,
		Kind:     "webhook",
		Object:   "stripe",
		Ref:      reqid,
		Op:       "call",
		Old:      nil,
		New:      vals,
		Metadata: nil,
	})
	if err != nil {
		log.WithError(err).Error("saving changelog")
	}

	switch event.Type {
	case "checkout.session.completed":
		err = srv.handleWebhookCheckoutSessionCompleted(ctx, &event)
	case "checkout.session.async_payment_succeeded":
		err = srv.handleWebhookCheckoutSessionSucceeded(ctx, &event)
	case "checkout.session.async_payment_failed":
		err = srv.handleWebhookCheckoutSessionFailed(ctx, &event)
	case "mandate.updated":
		go func() {
			err = srv.handleWebhookMandateUpdated(context.Background(), &event)
			if err != nil {
				log.WithError(err).Infof("HandleWebhook error on event : %+v\n", event)
			}
		}()
	case "payment_method.automatically_updated":
		err = srv.handleWebhookPaymentMethodAutoUpdate(ctx, &event)
	case "charge.succeeded":
		err = srv.handleWebhookChargeStatus(ctx, &event, cardpayment.PaymentStatus_SUCCEEDED)
	case "charge.failed":
		err = srv.handleWebhookChargeStatus(ctx, &event, cardpayment.PaymentStatus_FAILED)
	case "payment_intent.succeeded":
		err = srv.handleWebhookPaymentIntentStatus(ctx, &event, cardpayment.PaymentStatus_SUCCEEDED)
	case "payment_intent.created", "payment_intent.processing":
		err = srv.handleWebhookPaymentIntentStatus(ctx, &event, cardpayment.PaymentStatus_PROCESSING)
	case "payment_intent.payment_failed":
		err = srv.handleWebhookPaymentIntentStatus(ctx, &event, cardpayment.PaymentStatus_FAILED)
	case "invoice.payment_succeeded":
		err = srv.handleWebhookInvoiceStatus(ctx, &event, cardpayment.PaymentStatus_SUCCEEDED)
	case "invoice.payment_failed":
		err = srv.handleWebhookInvoiceStatus(ctx, &event, cardpayment.PaymentStatus_FAILED)
	case "customer.subscription.deleted":
		err = srv.handleWebhookCustomerSubscriptionStatus(ctx, &event, cardpayment.PaymentStatus_CANCELLED)
	}

	if err != nil {
		log.WithError(err).Infof("HandleWebhook error on event : %+v\n", event)

		return &cardpayment.HandleWebhookResponse{
			Success: false,
			Status:  cardpayment.HttpStatusCode_BAD_REQUEST,
		}, err
	}

	return &cardpayment.HandleWebhookResponse{
		Success: true,
		Status:  cardpayment.HttpStatusCode_OK,
	}, nil
}

func (srv *serviceImpl) HandleCheckout(
	ctx context.Context,
	request *cardpayment.HandleCheckoutRequest,
) (*cardpayment.HandleCheckoutResponse, error) {
	if request.SessionId == "" {
		return &cardpayment.HandleCheckoutResponse{
			Success: false,
			Status:  cardpayment.HttpStatusCode_BAD_REQUEST,
		}, nil
	}

	ch, er := srv.config.Repository.ChargeGetBySessionID(ctx, request.SessionId)
	if er != nil {
		return &cardpayment.HandleCheckoutResponse{
			Success: false,
			Status:  cardpayment.HttpStatusCode_BAD_REQUEST,
		}, er
	}

	if ch == nil {
		return &cardpayment.HandleCheckoutResponse{
			Success: false,
			Status:  cardpayment.HttpStatusCode_BAD_REQUEST,
		}, nil
	}

	m, cnt, err := srv.config.Repository.FilterMandates(ctx, &query.QueryParams{
		Filter: []*query.QueryParamsFilter{
			{
				Fields:   []string{"session_id"},
				Value:    []string{request.SessionId},
				Operator: query.QueryFilterOperator_EQUALS,
			},
		},
	})
	if err != nil {
		return &cardpayment.HandleCheckoutResponse{
			Property: ch.Property,
			Success:  false,
			Status:   cardpayment.HttpStatusCode_BAD_REQUEST,
		}, err
	}

	if cnt == 0 {
		return &cardpayment.HandleCheckoutResponse{
			Success:  true,
			Property: ch.Property,
			Status:   cardpayment.HttpStatusCode_OK,
		}, nil
	}

	if !request.Success {
		m[0].Status = cardpayment.MandateStatus_INACTIVE
		m[0].Valid = false

		err = srv.config.Repository.MandateSave(ctx, m[0])
		if err != nil {
			return &cardpayment.HandleCheckoutResponse{
				Success:  true,
				Property: ch.Property,
				Status:   cardpayment.HttpStatusCode_INTERNAL_SERVER_ERROR,
			}, nil
		}
	}

	return &cardpayment.HandleCheckoutResponse{
		Success:  true,
		Property: ch.Property,
		Status:   cardpayment.HttpStatusCode_OK,
	}, nil
}

func (srv *serviceImpl) handleWebhookInvoiceStatus(
	ctx context.Context,
	s *stripe.Event,
	status cardpayment.PaymentStatus,
) error {
	var invoice stripe.Invoice

	err := json.Unmarshal(s.Data.Raw, &invoice)
	if err != nil {
		return err
	}

	m, err := srv.config.Repository.ChargeGetByID(ctx, invoice.PaymentIntent.ID)
	if err != nil {
		return err
	}

	if status == cardpayment.PaymentStatus_SUCCEEDED {
		pi, _ := paymentintent.Get(
			invoice.PaymentIntent.ID,
			nil,
		)

		params := &stripe.SubscriptionParams{
			DefaultPaymentMethod: stripe.String(pi.PaymentMethod.ID),
		}

		_, err = sub.Update(invoice.Subscription.ID, params)
		if err != nil {
			return err
		}
	}

	m.Status = status

	err = srv.config.Repository.ChargeSave(ctx, m)
	if err != nil {
		return err
	}

	go srv.notifyUpdate(context.Background(), m)

	return nil
}

func (srv *serviceImpl) handleWebhookCustomerSubscriptionStatus(
	ctx context.Context,
	s *stripe.Event,
	status cardpayment.PaymentStatus,
) error {
	var subscription stripe.Subscription

	err := json.Unmarshal(s.Data.Raw, &subscription)
	if err != nil {
		return err
	}

	m, err := srv.config.Repository.ChargeGetBySubscriptionID(ctx, subscription.ID)
	if err != nil {
		return err
	}

	m.Status = status

	err = srv.config.Repository.ChargeSave(ctx, m)
	if err != nil {
		return err
	}

	go srv.notifyUpdate(context.Background(), m)

	return nil
}
