package service

import (
	"context"
	"encoding/json"
	"math"

	"github.com/stripe/stripe-go/v72/sub"

	"github.com/stripe/stripe-go/v72/paymentintent"
	"github.com/stripe/stripe-go/v72/transfer"
	cardpayment "gitlab.mashplace.com/mashplace/backend/cardpayment/api"
	"gitlab.mashplace.com/mashplace/backend/cardpayment/internal/model"
	"gitlab.mashplace.com/mashplace/backend/common/generic/grpc"
	"gitlab.mashplace.com/mashplace/backend/common/generic/log"
	extras "gitlab.mashplace.com/mashplace/backend/extras/api"
)

func (srv *serviceImpl) notifyClient(address string) (client cardpayment.CardpaymentNotifyServiceClient) {
	srv.notifyLock.RLock()

	client = srv.notifyClients[address]

	srv.notifyLock.RUnlock()

	if client != nil {
		return
	}

	srv.notifyLock.Lock()
	defer srv.notifyLock.Unlock()

	client = srv.notifyClients[address]

	if client != nil {
		return
	}

	client = cardpayment.NewCardpaymentNotifyServiceClient(grpc.Dial(address, "NotifyCardpayment", grpc.ClientCallback(
		func(ctx context.Context, req, resp interface{}, err error) {
			if !grpc.MethodIsMutation(ctx) {
				return
			}

			mdata := grpc.MethodMetadata(ctx)

			reqid := grpc.MethodRequestId(ctx)

			reqbs, _ := json.Marshal(req)

			var reqm map[string]interface{}

			_ = json.Unmarshal(reqbs, &reqm)

			fname := string(mdata.FullName())

			srv.RPCCallback(
				ctx,
				"grpc",
				"request",
				fname,
				reqid,
				reqm,
				nil,
			)

			if err == nil {
				var respm map[string]interface{}

				respbs, _ := json.Marshal(resp)

				_ = json.Unmarshal(respbs, &respm)

				srv.RPCCallback(
					ctx,
					"grpc",
					"response",
					fname,
					reqid,
					respm,
					nil,
				)
			} else {
				srv.RPCCallback(
					ctx,
					"grpc",
					"response",
					fname,
					reqid,
					map[string]interface{}{"error": err.Error()},
					nil,
				)
			}
		},
	)))

	srv.notifyClients[address] = client

	return
}

func (srv *serviceImpl) notifyUpdate(ctx context.Context, m *model.Charge) {
	if srv.config.ZeebeClient != nil {
		_, _ = srv.config.ZeebeClient.NewPublishMessageCommand().
			MessageName("payment-update").CorrelationKey(m.Property).Send(ctx)
	}

	if m.Status == cardpayment.PaymentStatus_SUCCEEDED {
		err := srv.processSuccess(ctx, m)
		if err != nil {
			log.WithError(err).Error("processing success change")
		}
	}

	if m.NotifyAddress != "" {
		var itemModels []*cardpayment.CardpaymentNotifyItem

		for _, item := range m.Items {
			itemModels = append(itemModels, &cardpayment.CardpaymentNotifyItem{
				Id:          item.ItemID,
				InsuranceId: item.TransferID,
			})
		}

		_, err := srv.notifyClient(m.NotifyAddress).CardpaymentNotify(ctx, &cardpayment.CardpaymentNotifyRequest{
			NotifyId: m.NotifyID,
			Status:   m.Status,
			Items:    itemModels,
			Details: &cardpayment.CardpaymentNotifyDetails{
				SubscriptionId:  m.SubscriptionID,
				SystemReference: m.ID,
			},
		})
		if err != nil {
			log.WithError(err).WithField("remote", m.NotifyAddress).Error("notifying remote")
		}
	}
}

func isServiceInsurance(serviceType extras.ServiceType) bool {
	switch serviceType {
	case extras.ServiceType_RENT_GUARANTEE_INSURANCE,
		extras.ServiceType_CONTENTS_INSURANCE:
		return true
	}

	return false
}

func (srv *serviceImpl) processSuccess(ctx context.Context, m *model.Charge) error {
	var amountInsurance float64

	var insuranceItems []*model.Service

	for _, item := range m.Items {
		if isServiceInsurance(item.Type) && item.TransferID == "" {
			insuranceItems = append(insuranceItems, item)
			amountInsurance += item.Price + item.TaxAmount
		}
	}

	var trInsuranceID string

	if len(insuranceItems) == 0 || m.Transferred {
		return nil
	}

	transferParamsInsurance := &stripe.TransferParams{
		Amount:      stripe.Int64(int64(math.Ceil(amountInsurance * 100))),
		Currency:    stripe.String(string(srv.config.Stripe.Currency)),
		Destination: stripe.String(srv.config.Stripe.InsuranceAccount),
	}

	if m.Type != cardpayment.PaymentType_SUBSCRIPTION {
		p, errP := paymentintent.Get(m.ID, nil)
		if errP != nil {
			return errP
		}

		transferParamsInsurance.SourceTransaction = stripe.String(p.Charges.Data[0].ID)
	} else {
		s, err := sub.Get(m.SubscriptionID, &stripe.SubscriptionParams{
			Params: stripe.Params{
				Expand: []*string{
					stripe.String("latest_invoice"),
				},
			},
		})
		if err != nil {
			return err
		}

		transferParamsInsurance.SourceTransaction = stripe.String(s.LatestInvoice.Charge.ID)
	}

	trInsurance, errTI := transfer.New(transferParamsInsurance)
	if errTI != nil {
		return errTI
	}

	trInsuranceID = trInsurance.ID

	for _, item := range insuranceItems {
		item.TransferID = trInsuranceID
	}

	m.Transferred = true

	err := srv.config.Repository.ChargeSave(ctx, m)
	if err != nil {
		return err
	}

	return nil
}
