package service

import (
	"context"
	"errors"
	"fmt"
	"math"
	"net/url"
	"sync"

	"github.com/stripe/stripe-go/v72/checkout/session"
	"github.com/stripe/stripe-go/v72/customer"
	"github.com/stripe/stripe-go/v72/mandate"
	"github.com/stripe/stripe-go/v72/paymentintent"
	"github.com/stripe/stripe-go/v72/price"
	"github.com/stripe/stripe-go/v72/product"
	"github.com/stripe/stripe-go/v72/sub"
	"go.mongodb.org/mongo-driver/bson/primitive"
	"go.mongodb.org/mongo-driver/mongo"
	"google.golang.org/protobuf/types/known/timestamppb"

	cardpayment "gitlab.mashplace.com/mashplace/backend/cardpayment/api"
	"gitlab.mashplace.com/mashplace/backend/cardpayment/internal/model"
	"gitlab.mashplace.com/mashplace/backend/common/api/generic/query"
	"gitlab.mashplace.com/mashplace/backend/common/api/grpcctx"
	"gitlab.mashplace.com/mashplace/backend/common/api/grpcerr"
	extras "gitlab.mashplace.com/mashplace/backend/extras/api"
)

var errNotOwned = grpcerr.New(grpcerr.AuthError, "not owned receipt")
var errCancelled = grpcerr.New(grpcerr.InvalidState, "receipt cancelled")
var errEmptyCustomerID = grpcerr.New(grpcerr.ProviderError, "empty customer ID")
var errInvalidDirectDebitType = grpcerr.New(grpcerr.ProviderError, "unsupported direct debit paymenet type")

type serviceImpl struct {
	config        Config
	notifyLock    sync.RWMutex
	notifyClients map[string]cardpayment.CardpaymentNotifyServiceClient
}

func (srv *serviceImpl) CreateDirectDebit(
	ctx context.Context,
	req *cardpayment.CreateDirectDebitRequest,
) (*cardpayment.CreateDirectDebitResponse, error) {
	if grpcctx.ContextIsTestUser(ctx) {
		return &cardpayment.CreateDirectDebitResponse{
			Success:   true,
			SessionId: "testsessionid",
			Id:        "testid",
		}, nil
	}

	switch req.GetType() {
	case cardpayment.PaymentType_DIRECT_DEBIT:
		return srv.directDebitPayment(ctx, req)
	case cardpayment.PaymentType_DIRECT_DEBIT_WITH_MANDATE:
		return srv.directDebitPaymentWithMandate(ctx, req)
	default:
		return nil, errInvalidDirectDebitType
	}
}

// directDebitPayment - creating checkout direct debit payment session with the mandate request,
// response include session ID for checkout
func (srv *serviceImpl) directDebitPayment(
	ctx context.Context,
	req *cardpayment.CreateDirectDebitRequest,
) (*cardpayment.CreateDirectDebitResponse, error) {
	if req.GetUserId() == "" {
		return nil, errors.New("user ID empty")
	}

	userMandate, err := srv.config.Repository.MandateGetByUserIDEmail(ctx, req.GetUserId(), req.GetUserEmail())
	if err != nil && err != mongo.ErrNoDocuments {
		return nil, err
	}

	var customerID string

	mandateExist := false

	if userMandate != nil {
		customerID = userMandate.CustomerID
		mandateExist = true
	} else {
		c, errm := srv.retrieveCustomer(ctx, req.GetUserId(), req.GetUserEmail(), req.GetUserName())
		if errm != nil {
			return nil, errm
		}
		customerID = c
	}

	offerID := req.GetCorrelationId()

	// this procedure necessary to found the offer for which the mandate comes in the webhook was ordered
	u, err := srv.config.Repository.UserOffersGet(ctx, offerID)
	if err != nil && err != mongo.ErrNoDocuments {
		return nil, err
	}

	if u == nil {
		u = &model.UserOffers{
			Offer: offerID,
			Users: []model.User{
				{
					ID:    req.GetUserId(),
					Email: req.GetUserEmail(),
				},
			},
		}
	} else {
		userAdded := false
		for _, user := range u.Users {
			if user.ID == req.GetUserId() {
				userAdded = true
				break
			}
		}
		if !userAdded {
			u.Users = append(u.Users, model.User{
				ID:    req.GetUserId(),
				Email: req.GetUserEmail(),
			})
		}
	}

	err = srv.config.Repository.UserOffersSave(ctx, u)

	if err != nil {
		return nil, err
	}

	amount, items := srv.createServiceAmount(req.GetItems())

	productObj, err := product.New(&stripe.ProductParams{
		Params: stripe.Params{
			Context: ctx,
		},
		Name: stripe.String(fmt.Sprintf("user %s for property %s", req.GetUserId(), req.GetProperty())),
	})
	if err != nil {
		return nil, err
	}

	priceObj, err := price.New(&stripe.PriceParams{
		Params: stripe.Params{
			Context: ctx,
		},
		Product:    stripe.String(productObj.ID),
		UnitAmount: stripe.Int64(int64(math.Ceil(amount * 100))),
		Currency:   stripe.String(string(srv.config.Stripe.Currency)),
	})
	if err != nil {
		return nil, err
	}

	s, err := session.New(&stripe.CheckoutSessionParams{
		Params: stripe.Params{
			Context: ctx,
			Metadata: map[string]string{
				"propertyId": req.GetProperty(),
				"userid":     req.GetUserId(),
				"invoiceId":  req.GetInvoiceId(),
			},
		},
		PaymentMethodTypes: stripe.StringSlice([]string{
			"bacs_debit",
		}),
		LineItems: []*stripe.CheckoutSessionLineItemParams{
			{
				Price:    stripe.String(priceObj.ID),
				Quantity: stripe.Int64(1),
			},
		},
		Mode:     stripe.String(string(stripe.CheckoutSessionModePayment)),
		Customer: stripe.String(customerID),
		PaymentIntentData: &stripe.CheckoutSessionPaymentIntentDataParams{
			Params: stripe.Params{
				Context: ctx,
			},
			SetupFutureUsage: stripe.String("off_session"),
		},
		SuccessURL: stripe.String(srv.config.Stripe.CallbackSuccessURL),
		CancelURL:  stripe.String(srv.config.Stripe.CallbackCancelURL),
	})
	if err != nil {
		return nil, err
	}

	m := &model.Charge{
		ID:            s.PaymentIntent.ID,
		SessionID:     s.ID,
		Items:         items,
		Type:          req.GetType(),
		UserID:        req.GetUserId(),
		UserEmail:     req.GetUserEmail(),
		Status:        cardpayment.PaymentStatus_PROCESSING,
		Property:      req.GetProperty(),
		NotifyID:      req.GetNotifyId(),
		NotifyAddress: req.GetNotifyAddress(),
		InvoiceNumber: req.GetInvoiceReference(),
		InvoiceID:     req.GetInvoiceId(),
	}

	err = srv.config.Repository.ChargeSave(ctx, m)
	if err != nil {
		return nil, err
	}

	if !mandateExist {
		md := &model.Mandate{
			SessionID:  s.ID,
			CustomerID: customerID,
			Offers:     []string{offerID},
			UserID:     req.GetUserId(),
			Email:      req.GetUserEmail(),
			Status:     cardpayment.MandateStatus_PENDING,
			Valid:      false,
		}

		err = srv.config.Repository.MandateSave(ctx, md)
		if err != nil {
			return nil, err
		}
	}

	return &cardpayment.CreateDirectDebitResponse{
		Success:   true,
		SessionId: s.ID,
		Id:        m.ID,
	}, nil
}

// directDebitPaymentWithMandate - creating checkout direct debit payment session using the user's mandate
func (srv *serviceImpl) directDebitPaymentWithMandate(
	ctx context.Context,
	req *cardpayment.CreateDirectDebitRequest,
) (*cardpayment.CreateDirectDebitResponse, error) {
	if req.GetUserId() == "" || req.GetUserEmail() == "" {
		return nil, fmt.Errorf("%w: userID and email are required", grpcerr.InvalidData)
	}

	userMandate, err := srv.config.Repository.MandateGetByUserIDEmail(ctx, req.GetUserId(), req.GetUserEmail())
	if err != nil {
		return nil, err
	}

	amount, items := srv.createServiceAmount(req.GetItems())

	intent, err := paymentintent.New(&stripe.PaymentIntentParams{
		Params: stripe.Params{
			Context: ctx,
			Metadata: map[string]string{
				"propertyId": req.GetProperty(),
				"userid":     req.GetUserId(),
				"invoiceId":  req.GetInvoiceId(),
			},
		},
		Customer:           stripe.String(userMandate.CustomerID),
		PaymentMethod:      stripe.String(userMandate.PaymentMethodID),
		Mandate:            stripe.String(userMandate.MandateID),
		Confirm:            stripe.Bool(true),
		PaymentMethodTypes: []*string{stripe.String("bacs_debit")},
		Amount:             stripe.Int64(int64(math.Ceil(amount * 100))),
		Currency:           stripe.String(string(srv.config.Stripe.Currency)),
		CaptureMethod:      stripe.String("automatic"),
		ReceiptEmail:       stripe.String(userMandate.Email),
	})
	if err != nil {
		return nil, err
	}

	m := &model.Charge{
		ID:            intent.ID,
		ClientSecret:  intent.ClientSecret,
		Items:         items,
		Type:          req.Type,
		UserID:        req.GetUserId(),
		UserEmail:     req.GetUserEmail(),
		Status:        cardpayment.PaymentStatus_PROCESSING,
		Property:      req.GetProperty(),
		NotifyID:      req.GetNotifyId(),
		NotifyAddress: req.GetNotifyAddress(),
		InvoiceNumber: req.GetInvoiceReference(),
		InvoiceID:     req.GetInvoiceId(),
	}

	err = srv.config.Repository.ChargeSave(ctx, m)
	if err != nil {
		return nil, err
	}

	return &cardpayment.CreateDirectDebitResponse{
		Success:   true,
		Id:        m.ID,
		SessionId: intent.ID,
	}, nil
}

func (srv *serviceImpl) createServiceAmount(svcs []*extras.ServiceSnapshot) (amount float64, items []*model.Service) {
	for _, e := range svcs {
		amount += e.GetPrice() + e.GetTaxAmount()

		items = append(items, &model.Service{
			ItemID:        e.GetId(),
			ServiceID:     e.GetServiceId(),
			Type:          e.GetType(),
			Code:          e.GetCode(),
			Price:         e.GetPrice(),
			TaxType:       e.GetTaxType(),
			TaxAmount:     e.GetTaxAmount(),
			TaxPercentage: e.GetTaxPercentage(),
			Title:         e.GetTitle(),
			Description:   e.GetDescription(),
			Primary:       e.GetPrimary(),
			Created:       e.GetCreated().AsTime(),
			Updated:       e.GetUpdated().AsTime(),
			Role:          e.GetRole(),
		})
	}

	return
}

func (srv *serviceImpl) PaymentWithMandate(
	ctx context.Context,
	req *cardpayment.CreateDirectDebitRequest,
) (*cardpayment.CreateDirectDebitResponse, error) {
	if grpcctx.ContextIsTestUser(ctx) {
		return &cardpayment.CreateDirectDebitResponse{
			Success:   true,
			SessionId: "testsessionid",
			Id:        "testid",
		}, nil
	}

	return srv.directDebitPaymentWithMandate(ctx, req)
}

func (srv *serviceImpl) retrieveCustomer(ctx context.Context, userID, email, name string) (string, error) {
	customerList := customer.List(&stripe.CustomerListParams{
		ListParams: stripe.ListParams{
			Context: ctx,
		},
		Email: stripe.String(email),
	})

	if customerList != nil {
		for _, c := range customerList.CustomerList().Data {
			if c.Description == userID {
				return c.ID, nil
			}
		}
	}

	c, err := customer.New(&stripe.CustomerParams{
		Params: stripe.Params{
			Context: ctx,
		},
		Name:        stripe.String(name),
		Email:       stripe.String(email),
		Description: stripe.String(userID),
	})
	if err != nil {
		return "", err
	}

	return c.ID, nil
}

func (srv *serviceImpl) createMandate(
	ctx context.Context,
	m *model.Mandate,
	email, user string,
) (err error) {
	if m == nil || m.Email != email || m.UserID != user {
		return
	}

	if m.Status != cardpayment.MandateStatus_ACTIVE {
		mResp, errm := mandate.Get(m.MandateID, nil)
		if errm != nil {
			return fmt.Errorf(
				"%w: error get mandate by ID: %s",
				grpcerr.DatabaseError,
				m.MandateID,
			)
		}

		if mResp.Status == stripe.MandateStatusActive {
			m.Status = cardpayment.MandateStatus_ACTIVE
			m.Valid = true
		} else if mResp.Status == stripe.MandateStatusInactive {
			m.Status = cardpayment.MandateStatus_INACTIVE
			m.Valid = false
		}
	}

	err = srv.config.Repository.MandateRemove(ctx, m.MandateID)
	if err != nil {
		return fmt.Errorf(
			"%w: error remove mandate %s",
			grpcerr.DatabaseError,
			m.MandateID,
		)
	}

	return
}

func (srv *serviceImpl) CreateMandate(
	ctx context.Context,
	req *cardpayment.CreateMandateRequest,
) (*cardpayment.CreateMandateResponse, error) {
	if grpcctx.ContextIsTestUser(ctx) {
		return &cardpayment.CreateMandateResponse{
			Success:   true,
			SessionId: "testsessionid",
			Created:   true,
		}, nil
	}

	if req.GetUserId() == "" || req.GetEmail() == "" {
		return nil, fmt.Errorf("%w: userID and email are required", grpcerr.InvalidData)
	}

	if req.GetOfferId() == "" {
		return nil, fmt.Errorf("%w: offerID are required", grpcerr.InvalidData)
	}

	m, err := srv.ensureMandate(ctx, req.GetUserId(), req.GetEmail(), "", req.GetOfferId())
	if err != nil {
		return nil, err
	}

	if len(m.Offers) == 0 {
		m.Offers = append(m.Offers, req.GetOfferId())

		err = srv.config.Repository.MandateSave(ctx, m)
		if err != nil {
			return nil, err
		}
	}

	return &cardpayment.CreateMandateResponse{
		Success:   true,
		SessionId: m.SessionID,
	}, nil
}

func (srv *serviceImpl) FilterMandates(
	ctx context.Context,
	req *cardpayment.FilterMandatesRequest,
) (*cardpayment.FilterMandatesResponse, error) {
	params := req.GetParams()
	if params == nil {
		params = &query.QueryParams{}
	}

	mandates, count, err := srv.config.Repository.FilterMandates(ctx, params)
	if err != nil {
		return nil, err
	}

	var res []*cardpayment.Mandate
	for _, m := range mandates {
		res = append(res, m.RenderMandate())
	}

	return &cardpayment.FilterMandatesResponse{
		Items: res,
		Total: int32(count),
	}, nil
}

func (srv *serviceImpl) GetMandate(
	ctx context.Context,
	req *cardpayment.GetMandateRequest,
) (*cardpayment.GetMandateResponse, error) {
	id, err := primitive.ObjectIDFromHex(req.GetMandateId())
	if err != nil {
		return nil, err
	}

	m, err := srv.config.Repository.MandateGetByID(ctx, id)
	if err != nil {
		return nil, err
	}

	return &cardpayment.GetMandateResponse{
		Mandate: m.RenderMandate(),
	}, nil
}

func (srv *serviceImpl) FormReceipt(
	ctx context.Context,
	req *cardpayment.FormReceiptRequest,
) (*cardpayment.FormReceiptResponse, error) {
	if grpcctx.ContextIsTestUser(ctx) {
		return &cardpayment.FormReceiptResponse{
			Id:           "testid",
			ClientSecret: "testclientsecret",
		}, nil
	}

	amount, items := srv.createServiceAmount(req.GetItems())

	paymentType := req.GetType()
	if paymentType == cardpayment.PaymentType_UNKNOWN_PAYMENT_TYPE {
		paymentType = cardpayment.PaymentType_CARD
	}

	var status cardpayment.PaymentStatus

	var id, clientSecret string

	switch {
	case amount == 0:
		status = cardpayment.PaymentStatus_SUCCEEDED
		id = "null_" + primitive.NewObjectID().Hex()
	case paymentType == cardpayment.PaymentType_BANK_ACCOUNT:
		status = cardpayment.PaymentStatus_SUCCEEDED
		id = "bank_" + primitive.NewObjectID().Hex()
	default:
		status = cardpayment.PaymentStatus_PROCESSING

		if amount < 0.5 {
			return nil, grpcerr.Newf(grpcerr.InvalidData, "too small amount to charge (%.2f < 0.50)", amount)
		}

		intent, err := paymentintent.New(&stripe.PaymentIntentParams{
			Params: stripe.Params{
				Context: ctx,
				Metadata: map[string]string{
					"propertyId": req.GetProperty(),
					"userid":     req.GetUserId(),
					"invoiceId":  req.GetInvoiceId(),
				},
			},
			Amount:             stripe.Int64(int64(math.Ceil(amount * 100))),
			Currency:           stripe.String(string(srv.config.Stripe.Currency)),
			CaptureMethod:      stripe.String("automatic"),
			PaymentMethodTypes: []*string{stripe.String("card")},
			ReceiptEmail:       stripe.String(req.GetUserEmail()),
		})
		if err != nil {
			return nil, err
		}

		id, clientSecret = intent.ID, intent.ClientSecret
	}

	m := &model.Charge{
		ID:            id,
		ClientSecret:  clientSecret,
		Items:         items,
		Type:          paymentType,
		UserID:        req.GetUserId(),
		UserEmail:     req.GetUserEmail(),
		Status:        status,
		Property:      req.GetProperty(),
		NotifyID:      req.GetNotifyId(),
		NotifyAddress: req.GetNotifyAddress(),
		InvoiceNumber: req.GetInvoiceReference(),
		InvoiceID:     req.GetInvoiceId(),
	}

	err := srv.config.Repository.ChargeSave(ctx, m)
	if err != nil {
		return nil, err
	}

	return &cardpayment.FormReceiptResponse{
		Id:           m.ID,
		ClientSecret: m.ClientSecret,
	}, nil
}

func (srv *serviceImpl) CheckReceipt(
	ctx context.Context,
	req *cardpayment.CheckReceiptRequest,
) (*cardpayment.CheckReceiptResponse, error) {
	m, err := srv.config.Repository.ChargeGetByID(ctx, req.GetId())
	if err != nil {
		return nil, err
	}

	if m.UserID != req.GetUserId() {
		return nil, errNotOwned
	}

	if m.Status == cardpayment.PaymentStatus_CANCELLED {
		return nil, errCancelled
	}

	intent, err := paymentintent.Get(req.GetId(), &stripe.PaymentIntentParams{
		Params: stripe.Params{
			Context: ctx,
		},
	})
	if err != nil {
		return nil, err
	}

	if intent.Status == stripe.PaymentIntentStatusSucceeded && m.Status != cardpayment.PaymentStatus_SUCCEEDED {
		m.Status = cardpayment.PaymentStatus_SUCCEEDED

		err = srv.config.Repository.ChargeSave(ctx, m)
		if err != nil {
			return nil, err
		}

		go srv.notifyUpdate(context.Background(), m)
	}

	return &cardpayment.CheckReceiptResponse{
		Receipt: renderCharge(m),
	}, nil
}

func (srv *serviceImpl) CancelReceipt(
	ctx context.Context,
	req *cardpayment.CancelReceiptRequest,
) (*cardpayment.CancelReceiptResponse, error) {
	m, err := srv.config.Repository.ChargeGetByID(ctx, req.GetId())
	if err != nil {
		return nil, err
	}

	if m.UserID != req.GetUserId() {
		return nil, errNotOwned
	}

	m.Status = cardpayment.PaymentStatus_CANCELLED

	err = srv.config.Repository.ChargeSave(ctx, m)
	if err != nil {
		return nil, err
	}

	_, err = paymentintent.Cancel(m.ID, &stripe.PaymentIntentCancelParams{
		Params: stripe.Params{
			Context: ctx,
		},
		CancellationReason: stripe.String(string(stripe.PaymentIntentCancellationReasonRequestedByCustomer)),
	})
	if err != nil {
		serr, ok := err.(*stripe.Error)
		if ok {
			return nil, grpcerr.New(grpcerr.ProviderError, serr.Msg)
		}

		return nil, grpcerr.New(grpcerr.ProviderError, err.Error())
	}

	return &cardpayment.CancelReceiptResponse{
		Success: true,
	}, nil
}

func (srv *serviceImpl) MarkReceiptPaid(
	ctx context.Context,
	req *cardpayment.MarkReceiptPaidRequest,
) (*cardpayment.MarkReceiptPaidResponse, error) {
	if grpcctx.ContextIsTestUser(ctx) {
		return &cardpayment.MarkReceiptPaidResponse{
			Success: true,
		}, nil
	}

	m, err := srv.config.Repository.ChargeGetByID(ctx, req.GetId())
	if err != nil {
		return nil, err
	}

	if m.UserID != req.GetUserId() {
		return nil, errNotOwned
	}

	if m.Status == cardpayment.PaymentStatus_CANCELLED {
		return nil, errCancelled
	}

	intent, err := paymentintent.Get(req.GetId(), &stripe.PaymentIntentParams{
		Params: stripe.Params{
			Context: ctx,
		},
	})
	if err != nil {
		return nil, err
	}

	if intent.Status == stripe.PaymentIntentStatusSucceeded && m.Status != cardpayment.PaymentStatus_SUCCEEDED {
		m.Status = cardpayment.PaymentStatus_SUCCEEDED

		err = srv.config.Repository.ChargeSave(ctx, m)
		if err != nil {
			return nil, err
		}

		go srv.notifyUpdate(context.Background(), m)
	}

	return &cardpayment.MarkReceiptPaidResponse{Success: true}, nil
}

func (srv *serviceImpl) ServiceTransferReceipt(
	ctx context.Context,
	req *cardpayment.ServiceTransferReceiptRequest,
) (*cardpayment.ServiceTransferReceiptResponse, error) {
	m, err := srv.config.Repository.ChargeGetByID(ctx, req.GetId())
	if err != nil {
		return nil, err
	}

	m.InvoiceID = req.GetInvoiceId()
	m.InvoiceNumber = req.GetInvoiceReference()
	m.NotifyAddress = req.GetNotifyAddress()
	m.NotifyID = req.GetNotifyId()

	err = srv.config.Repository.ChargeSave(ctx, m)
	if err != nil {
		return nil, err
	}

	var items []*extras.ServiceSnapshotStatus

	for _, i := range m.Items {
		svcCreatedAt := timestamppb.New(i.Created)
		svcUpdatedAt := timestamppb.New(i.Updated)

		items = append(items, &extras.ServiceSnapshotStatus{
			Service: &extras.ServiceSnapshot{
				Id:            i.ItemID,
				ServiceId:     i.ServiceID,
				Created:       svcCreatedAt,
				Updated:       svcUpdatedAt,
				Code:          i.Code,
				Type:          i.Type,
				Role:          i.Role,
				Price:         i.Price,
				TaxType:       i.TaxType,
				TaxAmount:     i.TaxAmount,
				TaxPercentage: i.TaxPercentage,
				Title:         i.Title,
				Description:   i.Description,
				Primary:       i.Primary,
			},
			UpdatedAt: svcUpdatedAt,
			Status:    extras.ServiceStatus_PENDING,
		})
	}

	return &cardpayment.ServiceTransferReceiptResponse{
		Type:         m.Type,
		ClientSecret: m.ClientSecret,
		UserId:       m.UserID,
		UserEmail:    m.UserEmail,
		Items:        items,
		Property:     m.Property,
		Status:       m.Status,
	}, nil
}

func (srv *serviceImpl) CreateSubscription(
	ctx context.Context,
	req *cardpayment.CreateSubscriptionRequest,
) (*cardpayment.CreateSubscriptionResponse, error) {
	if grpcctx.ContextIsTestUser(ctx) {
		return &cardpayment.CreateSubscriptionResponse{
			SessionId: "testsessionid",
			Url:       "https://example.com?testuser=testpayment",
		}, nil
	}

	customerID, err := srv.retrieveCustomer(ctx, req.GetUserId(), req.GetUserEmail(), req.GetUserName())
	if err != nil {
		return nil, err
	}

	var items []*stripe.SubscriptionItemsParams

	recur := &stripe.PriceRecurringParams{}

	switch req.GetInterval() {
	case cardpayment.Interval_INTERVAL_DAY:
		recur.Interval = stripe.String(string(stripe.PriceRecurringIntervalDay))
	case cardpayment.Interval_INTERVAL_WEEK:
		recur.Interval = stripe.String(string(stripe.PriceRecurringIntervalWeek))
	case cardpayment.Interval_INTERVAL_MONTH:
		recur.Interval = stripe.String(string(stripe.PriceRecurringIntervalMonth))
	case cardpayment.Interval_INTERVAL_YEAR:
		recur.Interval = stripe.String(string(stripe.PriceRecurringIntervalYear))
	}

	recur.IntervalCount = stripe.Int64(int64(req.GetIntervalCount()))

	for _, p := range req.GetItems() {
		var productObj *stripe.Product

		productObj, err = product.New(&stripe.ProductParams{
			Params: stripe.Params{
				Context: ctx,
			},
			Description: stripe.String(fmt.Sprintf("%s - %s for %s", p.GetDescription(), req.GetUserId(), req.GetProperty())),
			Name:        stripe.String(p.GetTitle()),
		})
		if err != nil {
			return nil, err
		}

		var priceObj *stripe.Price

		priceObj, err = price.New(&stripe.PriceParams{
			Params: stripe.Params{
				Context: ctx,
			},
			Currency:   stripe.String(string(srv.config.Stripe.Currency)),
			Product:    stripe.String(productObj.ID),
			Recurring:  recur,
			UnitAmount: stripe.Int64(int64(math.Ceil((p.GetPrice() + p.GetTaxAmount()) * 100))),
		})
		if err != nil {
			return nil, err
		}

		items = append(items, &stripe.SubscriptionItemsParams{
			Price:    stripe.String(priceObj.ID),
			Quantity: stripe.Int64(1),
		})
	}

	success, err := url.Parse(srv.config.Stripe.CallbackSuccessURL)
	if err != nil {
		return nil, err
	}

	cancel, err := url.Parse(srv.config.Stripe.CallbackCancelURL)
	if err != nil {
		return nil, err
	}

	if req.GetRedirectSuccess() != "" {
		q := success.Query()

		q.Set("redirect", req.GetRedirectSuccess())

		success.RawQuery = q.Encode()
	}

	if req.GetRedirectCancel() != "" {
		q := cancel.Query()

		q.Set("redirect", req.GetRedirectCancel())

		cancel.RawQuery = q.Encode()
	}

	sparams := &stripe.SubscriptionParams{
		Params: stripe.Params{
			Context: ctx,
			Metadata: map[string]string{
				"propertyId": req.GetProperty(),
				"userid":     req.GetUserId(),
				"invoiceId":  req.GetInvoiceId(),
			},
		},
		Customer:        stripe.String(customerID),
		Items:           items,
		PaymentBehavior: stripe.String("default_incomplete"),
	}
	sparams.AddExpand("latest_invoice.payment_intent")

	s, err := sub.New(sparams)
	if err != nil {
		return nil, err
	}

	_, svcitems := srv.createServiceAmount(req.GetItems())

	m := &model.Charge{
		ID:             s.LatestInvoice.PaymentIntent.ID,
		SessionID:      s.LatestInvoice.PaymentIntent.ID,
		Items:          svcitems,
		Type:           req.GetType(),
		UserID:         req.GetUserId(),
		UserEmail:      req.GetUserEmail(),
		Status:         cardpayment.PaymentStatus_PROCESSING,
		Property:       req.GetProperty(),
		NotifyID:       req.GetNotifyId(),
		NotifyAddress:  req.GetNotifyAddress(),
		InvoiceNumber:  req.GetInvoiceReference(),
		InvoiceID:      req.GetInvoiceId(),
		SubscriptionID: s.ID,
		ClientSecret:   s.LatestInvoice.PaymentIntent.ClientSecret,
	}

	err = srv.config.Repository.ChargeSave(ctx, m)
	if err != nil {
		return nil, err
	}

	return &cardpayment.CreateSubscriptionResponse{
		SessionId:    m.SubscriptionID,
		ClientSecret: m.ClientSecret,
	}, nil
}

func (srv *serviceImpl) CancelSubscription(
	ctx context.Context,
	req *cardpayment.CancelSubscriptionRequest,
) (*cardpayment.CancelSubscriptionResponse, error) {
	_, err := sub.Cancel(req.GetSubscriptionId(), &stripe.SubscriptionCancelParams{
		Params: stripe.Params{
			Context: ctx,
		},
	})
	if err != nil {
		return nil, err
	}

	m, err := srv.config.Repository.ChargeGetBySubscriptionID(ctx, req.GetSubscriptionId())
	if err != nil {
		return nil, err
	}

	m.Status = cardpayment.PaymentStatus_CANCELLED

	err = srv.config.Repository.ChargeSave(ctx, m)
	if err != nil {
		return nil, err
	}

	return &cardpayment.CancelSubscriptionResponse{}, nil
}

func (srv *serviceImpl) ensureMandate(
	ctx context.Context,
	userID, email, name, offerID string,
) (*model.Mandate, error) {
	m, err := srv.config.Repository.MandateGetByUserIDEmail(ctx, userID, email)
	if err != nil {
		if err != mongo.ErrNoDocuments {
			return nil, err
		}
	}

	err = srv.createMandate(
		ctx,
		m,
		email,
		userID,
	)
	if err != nil {
		return nil, err
	}

	if m != nil && m.Status == cardpayment.MandateStatus_ACTIVE {
		err = srv.config.Repository.MandateSave(ctx, m)
		if err != nil {
			return nil, fmt.Errorf("%w: error save mandate %s", grpcerr.DatabaseError, m.MandateID)
		}

		return m, nil
	}

	customerID, err := srv.retrieveCustomer(ctx, userID, email, name)
	if err != nil {
		return nil, err
	}

	if customerID == "" {
		return nil, errEmptyCustomerID
	}

	s, err := session.New(&stripe.CheckoutSessionParams{
		Params: stripe.Params{
			Context: ctx,
			Metadata: map[string]string{
				"offerId": offerID,
				"userid":  userID,
			},
		},
		PaymentMethodTypes: stripe.StringSlice([]string{
			"bacs_debit",
		}),
		Mode:          stripe.String(string(stripe.CheckoutSessionModeSetup)),
		Customer:      stripe.String(customerID),
		CustomerEmail: stripe.String(email),
		SuccessURL:    stripe.String(srv.config.Stripe.CallbackSuccessURL),
		CancelURL:     stripe.String(srv.config.Stripe.CallbackCancelURL),
	})
	if err != nil {
		return nil, err
	}

	var setupIntentID, mandateID, paymentMethodID string

	if s.SetupIntent != nil {
		setupIntentID = s.SetupIntent.ID

		if s.SetupIntent.Mandate != nil {
			mandateID = s.SetupIntent.Mandate.ID

			if s.SetupIntent.Mandate.PaymentMethod != nil {
				paymentMethodID = s.SetupIntent.Mandate.PaymentMethod.ID
			}
		}
	}

	m = &model.Mandate{
		MandateID:       mandateID,
		SessionID:       s.ID,
		SetupIntentID:   setupIntentID,
		CustomerID:      customerID,
		PaymentMethodID: paymentMethodID,
		UserID:          userID,
		Email:           email,
		Status:          cardpayment.MandateStatus_PENDING,
		Valid:           false,
	}

	err = srv.config.Repository.MandateSave(ctx, m)
	if err != nil {
		return nil, fmt.Errorf("%w: error save mandate %s", grpcerr.InvalidData, m.MandateID)
	}

	return m, nil
}

func (srv *serviceImpl) ChargeFilter(
	ctx context.Context,
	req *cardpayment.ChargeFilterRequest,
) (*cardpayment.ChargeFilterResponse, error) {
	params := req.GetParams()
	if params == nil {
		params = &query.QueryParams{}
	}

	charges, count, err := srv.config.Repository.FilterCharges(ctx, params)
	if err != nil {
		return nil, err
	}

	var res []*cardpayment.Charge
	for _, m := range charges {
		res = append(res, m.RenderCharge())
	}

	return &cardpayment.ChargeFilterResponse{
		Items: res,
		Total: int32(count),
	}, nil
}
